import math
import re
import string
from datetime import datetime

"""
Collection of classes that comprise the row type objects
in a nacha file
"""


class AchError(Exception):
    pass


class Ach(object):
    """
    Base class for ACH record fields
    """

    def make_space(self, spaces=1):
        """
        Return string with x number of spaces
        Defaults to 1
        """
        space_string = ""

        for i in range(spaces):
            space_string += " "

        return space_string

    def make_right_justified(self, field, length):
        """
        Return string with x number of leading spaces depending on field length
        Routing numbers should be 9 digits long, so we technically only need 1
        leading space.
        """

        if len(field) != length:
            return field.rjust(length)
        else:
            return field

    def make_zero(self, zeros=1):
        """
        Return string with x number of zeros
        Defaults to 1
        """
        zero_string = ""

        for i in range(zeros):
            zero_string += "0"

        return zero_string

    def validate_alpha_numeric_field(self, field, length):
        """
        Validates alpha numeric fields for nacha files
        field: (str)
        length: (int)
        """
        str_length = str(length)

        match = re.match(r"([\w,\s,'.','(',')','(',')','&','\-']{1," + str_length + "})", field)

        if match:
            if len(match.group(1)) < length:
                field = match.group(1) + self.make_space(length - len(match.group(1)))
            else:
                field = match.group(1)
        else:
            raise AchError("field does not match alpha numeric criteria")

        return field.upper()

    def validate_numeric_field(self, field, length):
        """
        Validates numeric field and zero right-pads if not
        long enough.
        field (int|str)
        length (int)
        """
        field = str(field)

        if field.isdigit():
            if len(field) < length:
                field = self.make_zero(length - len(field)) + field
            elif len(field) > length:
                raise AchError("field can only be %s digits long" % length)
        else:
            raise AchError("field needs to be numeric characters only")

        return field

    def validate_binary_field(self, field):
        """
        Validates binary string field (either '1' or '0')
        """

        if field not in ["1", "0"]:
            raise AchError("filed not '1' or '0'")
        return field


class Header(Ach):
    """
    Creates our File Header record of the nacha file
    """

    record_type_code = "1"
    priority_code = "01"
    record_size = "094"
    blk_factor = "10"
    format_code = "1"
    filler = " "

    alpha_numeric_fields = [
        "immediate_dest",
        "immediate_org",
        "file_id_mod",
        "im_dest_name",
        "im_orgn_name",
        "reference_code",
        "file_crt_date",
        "file_crt_time",
    ]

    field_lengths = {
        "immediate_dest": 9,
        "immediate_org": 10,
        "file_id_mod": 1,
        "im_dest_name": 23,
        "im_orgn_name": 23,
        "reference_code": 8,
        "file_crt_date": 6,
        "file_crt_time": 4,
    }

    def __init__(
        self,
        immediate_dest="",
        immediate_org="",
        file_id_mod="A",
        im_dest_name="",
        im_orgn_name="",
        reference_code="",
    ):
        """
        Initializes all values needed for
        our header row
        """

        date = datetime.today()

        self.immediate_dest = self.make_right_justified(immediate_dest, 9)
        self.immediate_org = self.make_right_justified(immediate_org, 10)
        self.file_crt_date = date.strftime("%y%m%d")
        self.file_crt_time = date.strftime("%H%M")
        self.file_id_mod = self.validate_file_id_mod(file_id_mod)
        self.im_dest_name = self.validate_alpha_numeric_field(im_dest_name, 23)
        self.im_orgn_name = self.validate_alpha_numeric_field(im_orgn_name, 23)

        if reference_code != "":
            self.reference_code = self.validate_alpha_numeric_field(reference_code, 8)
        else:
            self.reference_code = self.make_space(8)

    def __setattr__(self, name, value):
        if name in self.alpha_numeric_fields:
            value = self.validate_alpha_numeric_field(value, self.field_lengths[name])
        elif name == "file_id_mod":
            value = self.validate_file_id_mod(value)
        else:
            raise AchError("%s not in alpha numeric field list" % name)

        super(Header, self).__setattr__(name, value)

    def validate_file_id_mod(self, file_id_mod):
        """
        Validates the file ID modifier. It has to be ascii_uppercase
        and one character in length
        """
        if file_id_mod not in string.ascii_uppercase and len(file_id_mod) != 1:
            raise AchError("Invalid file_id_mod")

        return file_id_mod

    def get_row(self):
        """
        returns concatenated string of all parameters in
        nacha file
        """
        return (
            self.record_type_code
            + self.priority_code
            + self.filler
            + self.immediate_dest
            + self.immediate_org
            + self.file_crt_date
            + self.file_crt_time
            + self.file_id_mod
            + self.record_size
            + self.blk_factor
            + self.format_code
            + self.im_dest_name
            + self.im_orgn_name
            + self.reference_code
        )

    def get_count(self):
        """
        Returns length of all parameters in nach
        file
        """
        return len(self.get_row())


class FileControl(Ach):
    """
    Comprises the control record for an ACH file
    Appears at the end of file
    """

    record_type_code = "9"

    numeric_fields = [
        "batch_count",
        "block_count",
        "entadd_count",
        "entry_hash",
        "debit_amount",
        "credit_amount",
    ]

    alpha_numeric_fields = [
        "reserved",
    ]

    field_lengths = {
        "batch_count": 6,
        "block_count": 6,
        "entadd_count": 8,
        "entry_hash": 10,
        "debit_amount": 12,
        "credit_amount": 12,
        "reserved": 39,
    }

    def __init__(
        self,
        batch_count,
        block_count,
        entadd_count,
        entry_hash,
        debit_amount,
        credit_amount,
    ):
        """
        Initializes all the values we need for our file control record
        """

        self.batch_count = self.validate_numeric_field(batch_count, 6)
        self.block_count = self.validate_numeric_field(block_count, 6)
        self.entadd_count = self.validate_numeric_field(entadd_count, 8)
        self.entry_hash = self.validate_numeric_field(entry_hash, 10)
        self.debit_amount = self.validate_numeric_field(debit_amount, 12)
        self.credit_amount = self.validate_numeric_field(credit_amount, 12)
        self.reserved = self.make_space(39)

    def __setattr__(self, name, value):
        if name in self.numeric_fields:
            value = self.validate_numeric_field(value, self.field_lengths[name])
        elif name in self.alpha_numeric_fields:
            value = self.validate_alpha_numeric_field(value, self.field_lengths[name])
        else:
            raise AchError("%s not in numeric field list" % name)

        super(FileControl, self).__setattr__(name, value)

    def get_row(self):

        return (
            self.record_type_code
            + self.batch_count
            + self.block_count
            + self.entadd_count
            + self.entry_hash
            + self.debit_amount
            + self.credit_amount
            + self.reserved
        )

    def get_count(self):
        return len(self.get_row())


class BatchHeader(Ach):

    record_type_code = "5"

    std_ent_cls_code_list = [
        "ARC",
        "PPD",
        "CTX",
        "POS",
        "WEB",
        "BOC",
        "TEL",
        "MTE",
        "SHR",
        "CCD",
        "CIE",
        "POP",
        "RCK",
    ]

    serv_cls_code_list = ["200", "220", "225"]

    numeric_fields = ["orig_dfi_id", "batch_id", "eff_ent_date", "serv_cls_code"]

    alpha_numeric_fields = [
        "company_name",
        "cmpy_dis_data",
        "company_id",
        "std_ent_cls_code",
        "entry_desc",
        "desc_date",
        "orig_stat_code",
        "settlement_date",
    ]

    field_lengths = {
        "serv_cls_code": 3,
        "company_name": 16,
        "cmpy_dis_data": 20,
        "company_id": 10,
        "std_ent_cls_code": 3,
        "entry_desc": 10,
        "desc_date": 6,
        "eff_ent_date": 6,
        "settlement_date": 3,
        "orig_stat_code": 1,
        "orig_dfi_id": 8,
        "batch_id": 7,
    }

    def __init__(
        self,
        serv_cls_code="220",
        company_name="",
        cmpy_dis_data="",
        company_id="",
        std_ent_cls_code="CCD",
        entry_desc="",
        desc_date="",
        eff_ent_date="",
        settlement_date="000",
        orig_stat_code="",
        orig_dfi_id="",
        batch_id="",
    ):
        """
        Initializes and validates the values for our Batch Header
        rows. We use 220 and PPD as the default values for serv_cls_code
        and std_ent_cls_code.
        """

        args = locals().copy()

        self.settlement_date = self.make_space(3)

        for key in args:
            if key == "self":
                continue

            if args[key] != "":
                self.__setattr__(key, args[key])

            elif key in self.numeric_fields:
                self.__setattr__(key, self.make_zero(self.field_lengths[key]))

            elif key in self.alpha_numeric_fields:
                self.__setattr__(key, self.make_space(self.field_lengths[key]))

    def __setattr__(self, name, value):

        if name in self.numeric_fields:
            if name == "serv_cls_code" and str(value) not in self.serv_cls_code_list:
                raise AchError("%s not in serv_cls_code_list" % value)

            value = self.validate_numeric_field(value, self.field_lengths[name])

        elif name in self.alpha_numeric_fields:
            if name == "std_ent_cls_code" and str(value) not in self.std_ent_cls_code_list:
                raise AchError("%s not in std_ent_cls_code_list" % value)

            value = self.validate_alpha_numeric_field(value, self.field_lengths[name])

        else:
            raise AchError("%s not in numeric or alpha numeric fields list" % name)

        super(BatchHeader, self).__setattr__(name, value)

    def get_row(self):

        return (
            self.record_type_code
            + self.serv_cls_code
            + self.company_name
            + self.cmpy_dis_data
            + self.company_id
            + self.std_ent_cls_code
            + self.entry_desc
            + self.desc_date
            + self.eff_ent_date
            + self.settlement_date
            + self.orig_stat_code
            + self.orig_dfi_id
            + self.batch_id
        )

    def get_count(self):
        return len(self.get_row())


class BatchControl(Ach):

    record_type_code = "8"

    numeric_fields = [
        "serv_cls_code",
        "entadd_count",
        "entry_hash",
        "debit_amount",
        "credit_amount",
        "orig_dfi_id",
        "batch_id",
    ]

    alpha_numeric_fields = ["company_id", "mesg_auth_code", "reserved"]

    field_lengths = {
        "serv_cls_code": 3,
        "entadd_count": 6,
        "entry_hash": 10,
        "debit_amount": 12,
        "credit_amount": 12,
        "company_id": 10,
        "mesg_auth_code": 19,
        "reserved": 6,
        "orig_dfi_id": 8,
        "batch_id": 7,
    }

    def __init__(
        self,
        serv_cls_code="220",
        entadd_count="",
        entry_hash="",
        debit_amount="",
        credit_amount="",
        company_id="",
        orig_dfi_id="",
        batch_id="",
        mesg_auth_code="",
    ):
        """
        Initializes and validates the batch control record
        """
        args = locals().copy()

        self.reserved = self.make_space(6)

        for key in args:
            if key == "self":
                continue

            if args[key] != "":
                if key == "debit_amount" or key == "credit_amount":
                    self.__setattr__(key, int(100 * args[key]))
                else:
                    self.__setattr__(key, args[key])

            elif key in self.numeric_fields:
                self.__setattr__(key, self.make_zero(self.field_lengths[key]))

            elif key in self.alpha_numeric_fields:
                self.__setattr__(key, self.make_space(self.field_lengths[key]))

    def __setattr__(self, name, value):
        if name in self.numeric_fields:
            value = self.validate_numeric_field(value, self.field_lengths[name])
        elif name in self.alpha_numeric_fields:
            value = self.validate_alpha_numeric_field(value, self.field_lengths[name])
        else:
            raise AchError("%s not in numeric_fields or alpha_numeric_fields" % name)

        super(BatchControl, self).__setattr__(name, value)

    def get_row(self):

        return (
            self.record_type_code
            + self.serv_cls_code
            + self.entadd_count
            + self.entry_hash
            + self.debit_amount
            + self.credit_amount
            + self.company_id
            + self.mesg_auth_code
            + self.reserved
            + self.orig_dfi_id
            + self.batch_id
        )

    def get_count(self):
        return len(self.get_row())


class EntryDetail(Ach):
    """
    Object represents a single Entry Detail record of an ACH file
    """

    record_type_code = "6"

    std_ent_cls_code_list = [
        "ARC",
        "PPD",
        "CTX",
        "POS",
        "WEB",
        "BOC",
        "TEL",
        "MTE",
        "SHR",
        "CCD",
        "CIE",
        "POP",
        "RCK",
    ]

    numeric_fields = [
        "transaction_code",
        "recv_dfi_id",
        "check_digit",
        "amount",
        "num_add_recs",
        "card_exp_date",
        "doc_ref_num",
        "ind_card_acct_num",
        "card_tr_typ_code_shr",
        "add_rec_ind",
        "trace_num",
    ]

    alpha_numeric_fields = [
        "dfi_acnt_num",
        "chk_serial_num",
        "ind_name",
        "disc_data",
        "id_number",
        "recv_cmpy_name",
        "terminal_city",
        "terminal_state",
        "reserved",
        "card_tr_typ_code_pos",
        "pmt_type_code",
    ]

    field_lengths = {
        "transaction_code": 2,
        "recv_dfi_id": [8, 9],
        "check_digit": 1,
        "dfi_acnt_num": 17,
        "amount": 10,
        "chk_serial_num": [
            9,  # POP
            15,
        ],  # ARC, BOC
        "ind_name": [
            15,  # CIE, MTE
            22,
        ],  # ARC, BOC, CCD, PPD, TEL, POP, POS, WEB
        "disc_data": 2,
        "id_number": 15,
        "ind_id": 22,
        "num_add_recs": 4,
        "recv_cmpy_name": 16,
        "reserved": 2,
        "terminal_city": 4,
        "terminal_state": 2,
        "card_tr_typ_code_pos": 2,
        "card_tr_typ_code_shr": 2,
        "card_exp_date": 4,
        "doc_ref_num": 11,
        "ind_card_acct_num": 22,
        "pmt_type_code": 2,
        "add_rec_ind": 1,
        "trace_num": 15,
    }

    def __init__(
        self,
        std_ent_cls_code="CCD",
        transaction_code="",
        recv_dfi_id="",
        check_digit="",
        amount="",
        num_add_recs="",
        card_exp_date="",
        doc_ref_num="",
        ind_card_acct_num="",
        card_tr_typ_code_shr="",
        card_tr_typ_code_pos="",
        trace_num="",
        dfi_acnt_num="",
        ind_name="",
        disc_data="",
        id_number="",
        recv_cmpy_name="",
        chk_serial_num="",
        terminal_city="",
        terminal_state="",
        pmt_type_code="",
        add_rec_ind="",
    ):
        """
        Initialize and validate the values in Entry Detail record
        """
        self.std_ent_cls_code = std_ent_cls_code
        self.reserved = self.make_space(2)

        fields = locals().copy()

        for key in fields:
            if key == "self":
                continue

            if fields[key] != "":
                self.__setattr__(key, fields[key])

            elif key in ["chk_serial_num", "ind_name"]:
                if self.std_ent_cls_code in ["CIE", "MTE", "POP"]:
                    self.__setattr__(key, self.make_space(self.field_lengths[key][0]))
                else:
                    self.__setattr__(key, self.make_space(self.field_lengths[key][1]))

            elif key in self.numeric_fields:
                if key == "recv_dfi_id":
                    self.__setattr__(key, self.make_zero(self.field_lengths[key][0]))
                else:
                    self.__setattr__(key, self.make_zero(self.field_lengths[key]))

            elif key in self.alpha_numeric_fields:
                self.__setattr__(key, self.make_space(self.field_lengths[key]))

    def __setattr__(self, name, value):
        """
        Overides the setattr method for the object. We do this so
        that we can validate the field as it gets assigned.
        """

        if name in self.alpha_numeric_fields:
            # Special handling for Indvidiual/Company name field
            if name == "ind_name" and self.std_ent_cls_code in ["CIE", "MTE"]:
                value = self.validate_alpha_numeric_field(value, self.field_lengths[name][0])
            elif name == "ind_name":
                value = self.validate_alpha_numeric_field(value, self.field_lengths[name][1])

            # Special handling for Check serial number field
            elif name == "chk_serial_num" and self.std_ent_cls_code_list == "POP":
                value = self.validate_alpha_numeric_field(value, self.field_lengths[name][0])
            elif name == "chk_serial_num":
                value = self.validate_alpha_numeric_field(value, self.field_lengths[name][1])

            # The rest
            else:
                value = self.validate_alpha_numeric_field(value, self.field_lengths[name])

        elif name in self.numeric_fields:
            if name == "recv_dfi_id":
                try:
                    # try 8 digits first
                    value = self.validate_numeric_field(value, self.field_lengths[name][0])
                except AchError:
                    # now try to validate it 9 instead
                    value = self.validate_numeric_field(value, self.field_lengths[name][1])
            else:
                value = self.validate_numeric_field(value, self.field_lengths[name])

        elif name == "std_ent_cls_code" and value in self.std_ent_cls_code_list:
            pass

        else:
            raise AchError("%s not in numeric_fields or alpha_numeric_fields" % name)

        super(EntryDetail, self).__setattr__(name, value)

    def get_row(self):

        ret_string = ""

        ret_string = self.record_type_code + self.transaction_code + self.recv_dfi_id

        if len(self.recv_dfi_id) < 9:
            ret_string += self.check_digit

        ret_string += self.dfi_acnt_num + self.amount

        if self.std_ent_cls_code in ["ARC", "BOC"]:
            ret_string += self.chk_serial_num + self.ind_name + self.disc_data

        elif self.std_ent_cls_code in ["PPD", "TEL", 'CCD']:
            ret_string += self.id_number + self.ind_name + self.disc_data

        elif self.std_ent_cls_code == "CIE":
            ret_string += self.ind_name + self.ind_id + self.disc_data

        elif self.std_ent_cls_code == "CTX":
            ret_string += self.id_number + self.num_add_recs + self.recv_cmpy_name + self.reserved + self.disc_data

        elif self.std_ent_cls_code == "MTE":
            ret_string += self.ind_name + self.ind_id + self.disc_data

        elif self.std_ent_cls_code == "POP":
            ret_string += (
                self.chk_serial_num + self.terminal_city + self.terminal_state + self.ind_name + self.disc_data
            )

        elif self.std_ent_cls_code == "POS":
            ret_string += self.id_number + self.ind_name + self.card_tr_typ_code_pos

        elif self.std_ent_cls_code == "SHR":
            ret_string += self.card_exp_date + self.doc_ref_num + self.ind_card_acct_num + self.card_tr_typ_code_shr

        elif self.std_ent_cls_code == "RCK":
            ret_string += self.chk_serial_num + self.ind_name + self.disc_data

        elif self.std_ent_cls_code == "WEB":
            ret_string += self.id_number + self.ind_name + self.pmt_type_code

        ret_string += self.add_rec_ind + self.trace_num

        return ret_string

    def get_count(self):
        return len(self.get_row())

    def calc_check_digit(self):

        multipliers = [3, 7, 1, 3, 7, 1, 3, 7]

        tmp_num = 0

        for num, mult in zip(list(self.recv_dfi_id), multipliers):
            tmp_num += int(num) * mult

        nearest_10 = math.ceil(tmp_num / 10.0)

        self.check_digit = int((nearest_10 * 10) - tmp_num)


class AddendaRecord(Ach):

    record_type_code = "7"
    addenda_type_code = "05"

    alpha_numeric_fields = [
        "trans_desc",
        "net_id_code",
        "term_id_code",
        "trans_serial_code",
        "terminal_loc",
        "terminal_city",
        "terminal_state",
        "ref_info_1",
        "ref_info_2",
        "pmt_rel_info",
        "auth_card_exp",
    ]

    numeric_fields = [
        "trans_date",
        "trans_time",
        "trace_num",
        "ent_det_seq_num",
        "add_seq_num",
    ]

    field_lengths = {
        "trans_desc": 7,
        "net_id_code": 3,
        "term_id_code": 6,
        "trans_serial_code": 6,
        "terminal_loc": 27,
        "terminal_city": 15,
        "terminal_state": 2,
        "ref_info_1": 7,
        "ref_info_2": 3,
        "pmt_rel_info": 80,
        "trans_date": 4,
        "trans_time": 6,
        "trace_num": 15,
        "ent_det_seq_num": 7,
        "auth_card_exp": 6,
        "add_seq_num": 4,
    }

    def __init__(
        self,
        std_ent_cls_code="PPD",
        trans_desc="",
        net_id_code="",
        term_id_code="",
        ref_info_1="",
        ref_info_2="",
        trans_serial_code="",
        trans_date="",
        trans_time="",
        terminal_loc="",
        terminal_city="",
        terminal_state="",
        trace_num="",
        auth_card_exp="",
        add_seq_num="",
        ent_det_seq_num="",
        pmt_rel_info="",
    ):
        """
        Initializes and validates values in entry addenda rows
        """

        fields = locals().copy()

        self.std_ent_cls_code = std_ent_cls_code

        for key in fields:

            if key == "self":
                continue

            if fields[key] != "":
                self.__setattr__(key, fields[key])

            elif key in self.numeric_fields:
                self.__setattr__(key, self.make_zero(self.field_lengths[key]))

            elif key in self.alpha_numeric_fields:
                self.__setattr__(key, self.make_space(self.field_lengths[key]))

    def __setattr__(self, name, value):

        if name in self.alpha_numeric_fields:
            value = self.validate_alpha_numeric_field(value, self.field_lengths[name])
        elif name in self.numeric_fields:
            value = self.validate_numeric_field(value, self.field_lengths[name])
        elif name == "std_ent_cls_code":
            pass
        else:
            raise AchError("%s not in numeric or alpha numeric fields" % value)

        super(AddendaRecord, self).__setattr__(name, value)

    def get_row(self):

        ret_string = ""

        ret_string += self.record_type_code + self.addenda_type_code

        if self.std_ent_cls_code == "MTE":
            ret_string += (
                self.trans_desc
                + self.net_id_code
                + self.term_id_code
                + self.trans_serial_code
                + self.trans_date
                + self.trans_time
                + self.terminal_loc
                + self.terminal_city
                + self.terminal_state
                + self.trace_num
            )

        elif self.std_ent_cls_code in ["POS", "SHR"]:
            ret_string += (
                self.ref_info_1
                + self.ref_info_2
                + self.term_id_code
                + self.trans_serial_code
                + self.trans_date
                + self.auth_card_exp
                + self.terminal_loc
                + self.terminal_city
                + self.terminal_state
                + self.trace_num
            )

        else:
            ret_string += self.pmt_rel_info + self.add_seq_num + self.ent_det_seq_num

        return ret_string

    def get_count(self):
        return len(self.get_row())


import csv
import json
from io import StringIO


class SECCode:
    IAT = "IAT"
    IAT_NOC = "COR"


class IATParserConfig:
    ENTRY_DETAIL_DEF = [
        {"field": "record_type_code", "pos": 0, "len": 1},
        {"field": "transaction_code", "pos": 1, "len": 2},
        {"field": "recv_dfi_id", "pos": 3, "len": 9},
        {"field": "num_addenda", "pos": 12, "len": 4},
        {"field": "reserved_one", "pos": 16, "len": 13},
        {"field": "amount", "pos": 29, "len": 10},
        {"field": "dfi_acnt_num", "pos": 39, "len": 35},
        {"field": "reserved_two", "pos": 74, "len": 2},
        {"field": "ofac_screen_ind", "pos": 76, "len": 1},
        {"field": "ofac_screen_ind_two", "pos": 77, "len": 1},
        {"field": "add_rec_ind", "pos": 78, "len": 1},
        {"field": "trace_number", "pos": 79, "len": 15},
    ]

    BATCH_HEADER_DEF = [
        {"field": "record_type_code", "pos": 0, "len": 1},
        {"field": "serv_cls_code", "pos": 1, "len": 3},
        {"field": "iat_indicator", "pos": 4, "len": 16},
        {"field": "foreign_exchange_indicator", "pos": 20, "len": 2},
        {"field": "foreign_exchange_ref_indicator", "pos": 22, "len": 1},
        {"field": "iso_country_code", "pos": 38, "len": 2},
        {"field": "orig_id", "pos": 40, "len": 10},
        {"field": "std_ent_cls_code", "pos": 50, "len": 3},
        {"field": "entry_desc", "pos": 53, "len": 10},
        {"field": "iso_orig_currency_code", "pos": 63, "len": 3},
        {"field": "iso_dest_currency_code", "pos": 66, "len": 3},
        {"field": "eff_ent_date", "pos": 69, "len": 6},
        {"field": "settlement_date", "pos": 75, "len": 3},
        {"field": "orig_stat_code", "pos": 78, "len": 1},
        {"field": "orig_dfi_id", "pos": 79, "len": 8},
        {"field": "batch_id", "pos": 87, "len": 7},
    ]

    ADDENDA_710_RECORD_DEF = [
        {"field": "record_type_code", "pos": 0, "len": 1},
        {"field": "addenda_type_code", "pos": 1, "len": 2},
        {"field": "transaction_type_code", "pos": 3, "len": 3},
        {"field": "foreign_amount", "pos": 6, "len": 18},
        {"field": "foreign_trace_number", "pos": 24, "len": 22},
        {"field": "rec_name", "pos": 46, "len": 35},
        {"field": "reserved", "pos": 81, "len": 6},
        {"field": "ent_det_seq_num", "pos": 87, "len": 7},
    ]

    ADDENDA_711_RECORD_DEF = [
        {"field": "record_type_code", "pos": 0, "len": 1},
        {"field": "addenda_type_code", "pos": 1, "len": 2},
        {"field": "orig_name", "pos": 3, "len": 35},
        {"field": "orig_street_address", "pos": 38, "len": 35},
        {"field": "reserved", "pos": 73, "len": 14},
        {"field": "ent_det_seq_num", "pos": 87, "len": 7},
    ]

    ADDENDA_712_RECORD_DEF = [
        {"field": "record_type_code", "pos": 0, "len": 1},
        {"field": "addenda_type_code", "pos": 1, "len": 2},
        {"field": "orig_city_state", "pos": 3, "len": 35},
        {"field": "orig_country_postal", "pos": 38, "len": 35},
        {"field": "reserved", "pos": 73, "len": 14},
        {"field": "ent_det_seq_num", "pos": 87, "len": 7},
    ]

    ADDENDA_713_RECORD_DEF = [
        {"field": "record_type_code", "pos": 0, "len": 1},
        {"field": "addenda_type_code", "pos": 1, "len": 2},
        {"field": "orig_dfi_name", "pos": 3, "len": 35},
        {"field": "orig_dfi_id_qualifier", "pos": 38, "len": 2},
        {"field": "orig_dfi_id", "pos": 40, "len": 34},
        {"field": "orig_dfi_branch_cc", "pos": 74, "len": 3},
        {"field": "reserved", "pos": 77, "len": 10},
        {"field": "ent_det_seq_num", "pos": 87, "len": 7},
    ]

    ADDENDA_714_RECORD_DEF = [
        {"field": "record_type_code", "pos": 0, "len": 1},
        {"field": "addenda_type_code", "pos": 1, "len": 2},
        {"field": "rec_dfi_name", "pos": 3, "len": 35},
        {"field": "rec_dfi_id_qualifier", "pos": 38, "len": 2},
        {"field": "rec_dfi_id", "pos": 40, "len": 34},
        {"field": "rec_dfi_branch_cc", "pos": 74, "len": 3},
        {"field": "reserved", "pos": 77, "len": 10},
        {"field": "ent_det_seq_num", "pos": 87, "len": 7},
    ]

    ADDENDA_715_RECORD_DEF = [
        {"field": "record_type_code", "pos": 0, "len": 1},
        {"field": "addenda_type_code", "pos": 1, "len": 2},
        {"field": "rec_id", "pos": 3, "len": 15},
        {"field": "rec_street_address", "pos": 18, "len": 35},
        {"field": "reserved", "pos": 53, "len": 34},
        {"field": "ent_det_seq_num", "pos": 87, "len": 7},
    ]

    ADDENDA_716_RECORD_DEF = [
        {"field": "record_type_code", "pos": 0, "len": 1},
        {"field": "addenda_type_code", "pos": 1, "len": 2},
        {"field": "rec_city_state", "pos": 3, "len": 35},
        {"field": "rec_country_postal", "pos": 38, "len": 35},
        {"field": "reserved", "pos": 73, "len": 14},
        {"field": "ent_det_seq_num", "pos": 87, "len": 7},
    ]

    ADDENDA_717_RECORD_DEF = [
        {
            "field": "record_type_code",
            "pos": 0,
            "len": 1,
        },
        {
            "field": "addenda_type_code",
            "pos": 1,
            "len": 2,
        },
        {
            "field": "pmt_rel_info",
            "pos": 3,
            "len": 80,
        },
        {
            "field": "add_seq_num",
            "pos": 83,
            "len": 4,
        },
        {
            "field": "ent_det_seq_num",
            "pos": 87,
            "len": 7,
        },
    ]
    ADDENDA_718_RECORD_DEF = [
        {"field": "record_type_code", "pos": 0, "len": 1},
        {"field": "addenda_type_code", "pos": 1, "len": 2},
        {"field": "correspondent_bank_name", "pos": 3, "len": 35},
        {"field": "correspondent_bank_id_qual", "pos": 38, "len": 2},
        {"field": "correspondent_bank_id", "pos": 40, "len": 34},
        {"field": "correspondent_bank_country", "pos": 74, "len": 3},
        {"field": "reserved", "pos": 77, "len": 6},
        {"field": "add_seq_num", "pos": 83, "len": 4},
        {"field": "ent_det_seq_num", "pos": 87, "len": 7},
    ]

    ADDENDA_TYPE_TO_RECORD_DEF = {
        "710": ADDENDA_710_RECORD_DEF,
        "711": ADDENDA_711_RECORD_DEF,
        "712": ADDENDA_712_RECORD_DEF,
        "713": ADDENDA_713_RECORD_DEF,
        "714": ADDENDA_714_RECORD_DEF,
        "715": ADDENDA_715_RECORD_DEF,
        "716": ADDENDA_716_RECORD_DEF,
        "717": ADDENDA_717_RECORD_DEF,
        "718": ADDENDA_718_RECORD_DEF,
    }


class Parser(IATParserConfig):
    """
    Parser for ACH files
    """

    FILE_HEADER = "1"
    FILE_CONTROL = "9"
    BATCH_HEADER = "5"
    BATCH_CONTROL = "8"
    ENTRY_DETAIL = "6"
    ADDENDA_RECORD = "7"

    FILE_HEADER_DEF = [
        {
            "field": "record_type_code",
            "pos": 0,
            "len": 1,
        },
        {
            "field": "priority_code",
            "pos": 1,
            "len": 2,
        },
        {
            "field": "filler",
            "pos": 3,
            "len": 1,
        },
        {
            "field": "immediate_dest",
            "pos": 4,
            "len": 9,
        },
        {
            "field": "immediate_org",
            "pos": 13,
            "len": 10,
        },
        {
            "field": "file_crt_date",
            "pos": 23,
            "len": 6,
        },
        {
            "field": "file_crt_time",
            "pos": 29,
            "len": 4,
        },
        {
            "field": "file_id_mod",
            "pos": 33,
            "len": 1,
        },
        {
            "field": "record_size",
            "pos": 34,
            "len": 3,
        },
        {
            "field": "blk_factor",
            "pos": 37,
            "len": 2,
        },
        {
            "field": "format_code",
            "pos": 39,
            "len": 1,
        },
        {
            "field": "im_dest_name ",
            "pos": 40,
            "len": 23,
        },
        {
            "field": "im_orgn_name ",
            "pos": 63,
            "len": 23,
        },
        {
            "field": "reference_code",
            "pos": 86,
            "len": 8,
        },
    ]

    FILE_CONTROL_DEF = [
        {
            "field": "record_type_code",
            "pos": 0,
            "len": 1,
        },
        {
            "field": "batch_count",
            "pos": 1,
            "len": 6,
        },
        {
            "field": "block_count",
            "pos": 7,
            "len": 6,
        },
        {
            "field": "entadd_count",
            "pos": 13,
            "len": 8,
        },
        {
            "field": "entry_hash",
            "pos": 21,
            "len": 10,
        },
        {
            "field": "debit_amount",
            "pos": 31,
            "len": 12,
        },
        {
            "field": "credit_amount",
            "pos": 43,
            "len": 12,
        },
        {
            "field": "reserved",
            "pos": 55,
            "len": 39,
        },
    ]

    BATCH_HEADER_DEF = [
        {
            "field": "record_type_code",
            "pos": 0,
            "len": 1,
        },
        {
            "field": "serv_cls_code",
            "pos": 1,
            "len": 3,
        },
        {
            "field": "company_name",
            "pos": 4,
            "len": 16,
        },
        {
            "field": "cmpy_dis_data",
            "pos": 20,
            "len": 20,
        },
        {
            "field": "company_id",
            "pos": 40,
            "len": 10,
        },
        {
            "field": "std_ent_cls_code",
            "pos": 50,
            "len": 3,
        },
        {
            "field": "entry_desc",
            "pos": 53,
            "len": 10,
        },
        {
            "field": "desc_date",
            "pos": 63,
            "len": 6,
        },
        {
            "field": "eff_ent_date",
            "pos": 69,
            "len": 6,
        },
        {
            "field": "settlement_date",
            "pos": 75,
            "len": 3,
        },
        {
            "field": "orig_stat_code",
            "pos": 78,
            "len": 1,
        },
        {
            "field": "orig_dfi_id",
            "pos": 79,
            "len": 8,
        },
        {
            "field": "batch_id",
            "pos": 87,
            "len": 7,
        },
    ]

    BATCH_CONTROL_DEF = [
        {
            "field": "record_type_code",
            "pos": 0,
            "len": 1,
        },
        {
            "field": "serv_cls_code",
            "pos": 1,
            "len": 3,
        },
        {
            "field": "entadd_count",
            "pos": 4,
            "len": 6,
        },
        {
            "field": "entry_hash",
            "pos": 10,
            "len": 10,
        },
        {
            "field": "debit_amount",
            "pos": 20,
            "len": 12,
        },
        {
            "field": "credit_amount",
            "pos": 32,
            "len": 12,
        },
        {
            "field": "company_id",
            "pos": 44,
            "len": 10,
        },
        {
            "field": "mesg_auth_code",
            "pos": 54,
            "len": 19,
        },
        {
            "field": "reserved",
            "pos": 73,
            "len": 6,
        },
        {
            "field": "orig_dfi_id",
            "pos": 79,
            "len": 8,
        },
        {
            "field": "orig_dfi_id",
            "pos": 87,
            "len": 7,
        },
    ]

    ENTRY_DETAIL_DEF = [
        {
            "field": "record_type_code",
            "pos": 0,
            "len": 1,
        },
        {
            "field": "transaction_code",
            "pos": 1,
            "len": 2,
        },
        {
            "field": "recv_dfi_id",
            "pos": 3,
            "len": 8,
        },
        {
            "field": "check_digit",
            "pos": 11,
            "len": 1,
        },
        {
            "field": "dfi_acnt_num",
            "pos": 12,
            "len": 17,
        },
        {
            "field": "amount",
            "pos": 29,
            "len": 10,
        },
        {
            "field": "ind_id",
            "pos": 39,
            "len": 15,
        },
        {
            "field": "ind_name",
            "pos": 54,
            "len": 22,
        },
        {
            "field": "disc_data",
            "pos": 76,
            "len": 2,
        },
        {
            "field": "add_rec_ind",
            "pos": 78,
            "len": 1,
        },
        {
            "field": "trace_num",
            "pos": 79,
            "len": 15,
        },
    ]

    ADDENDA_RECORD_DEF = [
        {
            "field": "record_type_code",
            "pos": 0,
            "len": 1,
        },
        {
            "field": "addenda_type_code",
            "pos": 1,
            "len": 2,
        },
        {
            "field": "pmt_rel_info",
            "pos": 3,
            "len": 80,
        },
        {
            "field": "add_seq_num",
            "pos": 83,
            "len": 4,
        },
        {
            "field": "ent_det_seq_num",
            "pos": 87,
            "len": 7,
        },
    ]

    record_type_codes = {
        "1": "file_header",
        "9": "file_control",
        "5": "batch_header",
        "8": "batch_control",
        "6": "entry_detail",
        "7": "addenda_record",
    }

    def __init__(self, ach_file):
        self.ach_file = ach_file
        self.ach_lines = ach_file.split("\n")
        self.ach_data = {}

        self.__parse_file()

    def as_json(self):
        return json.dumps(self.ach_data)

    def as_dict(self):
        return self.ach_data

    def as_csv(self):
        data = self.as_dict()
        buffer = StringIO()
        file_header = [f["field"] for f in self.FILE_HEADER_DEF]
        batch_header = [f["field"] for f in self.BATCH_HEADER_DEF]
        entry_header = [f["field"] for f in self.ENTRY_DETAIL_DEF]
        addenda_header = [f["field"] for f in self.ADDENDA_RECORD_DEF]
        batch_control = [f["field"] for f in self.BATCH_CONTROL_DEF]
        file_control = [f["field"] for f in self.FILE_CONTROL_DEF]

        writer = csv.DictWriter(buffer, fieldnames=file_header)
        writer.writeheader()
        writer.writerow(data["file_header"])
        writer.writerow({})
        for batch in data["batches"]:
            writer.fieldnames = batch_header
            writer.writeheader()
            writer.writerow(batch["batch_header"])
            max_addenda = 0
            for entry in batch["entries"]:
                max_addenda = max(max_addenda, len(entry["addenda"]))
            writer.writerow({})
            if max_addenda <= 1:
                writer.fieldnames = entry_header + ["a_{}".format(h) for h in addenda_header]
            else:
                writer.fieldnames = entry_header + [
                    "a_{}_{}".format(h, i) for h in addenda_header for i in range(max_addenda)
                ]
            writer.writeheader()
            for entry in batch["entries"]:
                detail = dict(entry["entry_detail"])
                if entry["addenda"]:
                    if len(entry["addenda"]) == 1:
                        detail.update({"a_{}".format(k): v for k, v in entry["addenda"][0].items()})
                    else:
                        for i, addenda in enumerate(entry["addenda"]):
                            detail.update({"a_{}_{}".format(k, i): v for k, v in addenda.items()})
                writer.writerow(detail)
            writer.writerow({})
            writer.fieldnames = batch_control
            writer.writeheader()
            writer.writerow(batch["batch_control"])
            writer.writerow({})
        writer.fieldnames = file_control
        writer.writeheader()
        writer.writerow(data["file_control"])
        return buffer.getvalue()

    def __parse_file(self):
        self.__parse_file_header()
        self.__parse_file_control()

        batch_info = self.__get_batch_info()
        self.__parse_batches(batch_info)

    def __parse_line(self, line, definitions):
        record_data = {}

        for rule in definitions:
            value = line[rule["pos"] : rule["pos"] + rule["len"]]
            record_data[rule["field"]] = value

        return record_data

    def __parse_file_header(self):
        for line in self.ach_lines:
            if line:
                if line[0] == self.FILE_HEADER:
                    self.ach_data["file_header"] = self.__parse_line(line, self.FILE_HEADER_DEF)
                    break

    def __parse_file_control(self):
        for line in self.ach_lines:
            if line:
                if line[0] == self.FILE_CONTROL:
                    self.ach_data["file_control"] = self.__parse_line(line, self.FILE_CONTROL_DEF)
                    break

    def __get_batch_info(self):
        batches = []

        for line_num, line in enumerate(self.ach_lines):
            if line:
                if line[0] == self.BATCH_HEADER:
                    batches.append(
                        {
                            "batch_header_line": line_num,
                        }
                    )
                if line[0] == self.BATCH_CONTROL:
                    batches[len(batches) - 1]["batch_control_line"] = line_num

        return batches

    def __parse_batches(self, batch_info):
        self.ach_data["batches"] = []

        for batch in batch_info:
            batch_header = self.__parse_line(self.ach_lines[batch["batch_header_line"]], self.BATCH_HEADER_DEF)

            sec_code = batch_header["std_ent_cls_code"]
            if sec_code == SECCode.IAT:
                batch_header = self.__parse_line(
                    self.ach_lines[batch["batch_header_line"]], IATParserConfig.BATCH_HEADER_DEF
                )

            self.ach_data["batches"].append(
                {
                    "batch_header": batch_header,
                    "batch_control": self.__parse_line(
                        self.ach_lines[batch["batch_control_line"]], self.BATCH_CONTROL_DEF
                    ),
                    "entries": [],
                }
            )

            start = batch["batch_header_line"] + 1
            stop = batch["batch_control_line"]

            for line_num in range(start, stop):
                if self.ach_lines[line_num]:
                    cur_batch = len(self.ach_data["batches"]) - 1
                    cur_entry = len(self.ach_data["batches"][cur_batch]["entries"]) - 1

                    if self.ach_lines[line_num][0] == self.ENTRY_DETAIL:
                        if sec_code == SECCode.IAT:
                            entry_detail = self.__parse_line(self.ach_lines[line_num], IATParserConfig.ENTRY_DETAIL_DEF)
                        else:
                            entry_detail = self.__parse_line(self.ach_lines[line_num], self.ENTRY_DETAIL_DEF)

                        self.ach_data["batches"][cur_batch]["entries"].append(
                            {"entry_detail": entry_detail, "addenda": []}
                        )
                    if self.ach_lines[line_num][0] == self.ADDENDA_RECORD:
                        if sec_code == SECCode.IAT:
                            addenda_type = self.ach_lines[line_num][:3]
                            definitions = IATParserConfig.ADDENDA_TYPE_TO_RECORD_DEF.get(addenda_type)
                            if definitions:
                                addenda_detail = self.__parse_line(self.ach_lines[line_num], definitions)
                        else:
                            addenda_detail = self.__parse_line(self.ach_lines[line_num], self.ADDENDA_RECORD_DEF)

                        self.ach_data["batches"][cur_batch]["entries"][cur_entry]["addenda"].append(addenda_detail)


import math
from datetime import datetime, timedelta

from pyACH.data_types import (
    Header,
    FileControl,
    BatchHeader,
    BatchControl,
    EntryDetail,
    AddendaRecord,
)


class AchFile(object):
    """
    This class is what stores the ach data.  Its main external methods
    are `add_batch` and `render_to_string`.
    """

    def __init__(self, file_id_mod, settings):
        """
        The file_id_mod should be 'A' for the first of the day, 'B'
        for the second and so on.
        """
        self.settings = settings

        try:
            company_name = settings.get(
                "company_name",
                settings["immediate_org_name"],
            )
            self.settings["company_name"] = company_name
            self.header = Header(
                settings["immediate_dest"],
                settings["immediate_org"],
                file_id_mod,
                settings["immediate_dest_name"],
                settings["immediate_org_name"],
            )

        except KeyError:
            raise Exception(
                'Settings require: "immediate_dest", "immediate_org", \
                immediate_dest_name", and "immediate_org_name"'
            )

        self.batches = list()

    def add_batch(
        self,
        std_ent_cls_code,
        batch_entries=None,
        credits=True,
        debits=False,
        eff_ent_date=None,
        company_id=None,
        entry_desc=None,
        company_name=None,
    ):
        """
        Use this to add batches to the file. For valid std_ent_cls_codes see:
        http://en.wikipedia.org/wiki/Automated_Clearing_House#SEC_codes
        """
        if batch_entries is None:
            batch_entries = list()

        if entry_desc is not None:
            entry_desc = entry_desc[:10].upper()
        entry_desc = entry_desc or self.get_entry_desc(std_ent_cls_code)

        batch_count = len(self.batches) + 1

        if not eff_ent_date:
            eff_ent_date = datetime.today() + timedelta(days=1)

        if credits and debits:
            serv_cls_code = "200"
        elif credits:
            serv_cls_code = "220"
        elif debits:
            serv_cls_code = "225"

        date = datetime.today()

        batch_header = BatchHeader(
            serv_cls_code=serv_cls_code,
            batch_id=batch_count,
            company_id=company_id or self.settings["company_id"],
            std_ent_cls_code=std_ent_cls_code,
            entry_desc=entry_desc,
            desc_date=date.strftime("%m%d%y"),
            eff_ent_date=eff_ent_date.strftime("%y%m%d"),  # YYMMDD
            settlement_date='000',
            orig_stat_code="1",
            orig_dfi_id=self.settings["immediate_dest"][:8],
            company_name=(company_name or self.settings["company_name"])[:16],
        )

        entries, failed_entry_errors = [], []
        entry_counter = 1

        for record in batch_entries:
            try:
                entry = EntryDetail(
                    std_ent_cls_code=std_ent_cls_code,
                    id_number=record.get("id_number", ""),
                )

                entry.transaction_code = record.get("type")
                entry.recv_dfi_id = record.get("routing_number")


                if len(record["routing_number"]) < 9:
                    entry.calc_check_digit()
                else:
                    entry.check_digit = record["routing_number"][8]

                entry.dfi_acnt_num = record["account_number"]

                entry.amount = int(round(float(record["amount"]) * 100))
                entry.ind_name = record["name"].upper()[:22]
                entry.trace_num = self.settings["trace_num"][:8] + entry.validate_numeric_field(entry_counter, 7)

                entries.append((entry, record.get("addenda", [])))
                entry_counter += 1
            except Exception as e:
                failed_entry_errors.append((record, e))

        self.batches.append(FileBatch(batch_header, entries))
        self.set_control()
        return failed_entry_errors

    def set_control(self):

        batch_count = len(self.batches)
        block_count = self.get_block_count(self.batches)
        entry_hash = self.get_entry_hash(self.batches)
        entadd_count = self.get_entadd_count(self.batches)
        debit_amount = self.get_debit_amount(self.batches)
        credit_amount = self.get_credit_amount(self.batches)

        self.control = FileControl(
            batch_count,
            block_count,
            entadd_count,
            entry_hash,
            debit_amount,
            credit_amount,
        )

    def get_block_count(self, batches):

        return int(math.ceil(self.get_lines(batches) / 10.0))

    def get_lines(self, batches):
        header_count = 1
        control_count = 1
        batch_header_count = len(batches)
        batch_footer_count = batch_header_count

        entadd_count = self.get_entadd_count(batches)

        lines = header_count + control_count + batch_header_count + batch_footer_count + entadd_count

        return lines

    def get_entadd_count(self, batches):
        entadd_count = 0

        for batch in batches:
            entadd_count = entadd_count + int(batch.batch_control.entadd_count)

        return entadd_count

    def get_entry_hash(self, batches):
        entry_hash = 0

        for batch in batches:
            entry_hash = entry_hash + int(batch.batch_control.entry_hash)

        if len(str(entry_hash)) > 10:
            pos = len(str(entry_hash)) - 10
            entry_hash = str(entry_hash)[pos:]
        else:
            entry_hash = str(entry_hash)

        return entry_hash

    def get_debit_amount(self, batches):
        debit_amount = 0

        for batch in batches:
            debit_amount = debit_amount + int(batch.batch_control.debit_amount)

        return debit_amount

    def get_credit_amount(self, batches):
        credit_amount = 0

        for batch in batches:
            credit_amount = credit_amount + int(batch.batch_control.credit_amount)

        return credit_amount

    def get_nines(self, rows, line_ending):
        nines = ""

        for i in range(rows):
            nines += "9" * 94
            if i == rows - 1:
                continue
            nines += line_ending

        return nines

    def get_entry_desc(self, std_ent_cls_code):

        if std_ent_cls_code == "PPD":
            entry_desc = "PAYROLL"
        elif std_ent_cls_code == "CCD":
            entry_desc = "PAYMENT"
        else:
            entry_desc = "OTHER"

        return entry_desc

    def render_to_string(self, force_crlf=False):
        """
        Renders a nacha file as a string
        """
        line_ending = "\n"
        if force_crlf:
            line_ending = "\r\n"

        ret_string = self.header.get_row() + line_ending

        for batch in self.batches:
            ret_string += batch.render_to_string(force_crlf=force_crlf)

        ret_string += self.control.get_row() + line_ending

        lines = self.get_lines(self.batches)

        nine_lines = int(round(10 * (math.ceil(lines / 10.0) - (lines / 10.0))))

        ret_string += self.get_nines(nine_lines, line_ending)

        return ret_string


class FileBatch(object):
    """
    Holds:
    BatchHeader  (1)
    Entry        (n) <-- multiple
    BatchControl (1)
    """

    def __init__(self, batch_header, entries):
        """
        args: batch_header (BatchHeader), entries (List[FileEntry])
        """

        entadd_count = 0

        self.batch_header = batch_header
        self.entries = []

        for entry, addenda in entries:
            entadd_count += 1
            entadd_count += len(addenda)
            self.entries.append(FileEntry(entry, addenda))

        # set up batch_control

        batch_control = BatchControl(self.batch_header.serv_cls_code)

        batch_control.entadd_count = entadd_count
        batch_control.entry_hash = self.get_entry_hash(self.entries)
        batch_control.debit_amount = self.get_debit_amount(self.entries)
        batch_control.credit_amount = self.get_credit_amount(self.entries)
        batch_control.company_id = self.batch_header.company_id
        batch_control.orig_dfi_id = self.batch_header.orig_dfi_id
        batch_control.batch_id = self.batch_header.batch_id

        self.batch_control = batch_control

    def get_entry_hash(self, entries):

        entry_hash = 0

        for entry in entries:
            entry_hash += int(entry.entry_detail.recv_dfi_id[:8])

        if len(str(entry_hash)) > 10:
            pos = len(str(entry_hash)) - 10
            entry_hash = str(entry_hash)[pos:]
        else:
            entry_hash = str(entry_hash)

        return entry_hash

    def get_debit_amount(self, entries):
        debit_amount = 0

        for entry in entries:
            if str(entry.entry_detail.transaction_code) in ["27", "37", "28", "38"]:
                debit_amount = debit_amount + int(entry.entry_detail.amount)

        return debit_amount

    def get_credit_amount(self, entries):
        credit_amount = 0

        for entry in entries:
            if str(entry.entry_detail.transaction_code) in ["22", "32", "23", "33"]:
                credit_amount += int(entry.entry_detail.amount)

        return credit_amount

    def render_to_string(self, force_crlf=False):
        """
        Renders a nacha file batch to string
        """
        line_ending = "\n"
        if force_crlf:
            line_ending = "\r\n"

        ret_string = self.batch_header.get_row() + line_ending

        for entry in self.entries:
            ret_string += entry.render_to_string(force_crlf=force_crlf)

        ret_string += self.batch_control.get_row() + line_ending

        return ret_string


class FileEntry(object):
    """
    Holds:
    EntryDetail (1)
    AddendaRecord (n) <-- for some types of entries there can be more than one
    """

    def __init__(self, entry_detail, addenda_record=[]):
        """
        args: entry_detail( EntryDetail), addenda_record (List[AddendaRecord])
        """

        self.entry_detail = entry_detail
        self.addenda_record = []

        for index, addenda in enumerate(addenda_record):
            self.addenda_record.append(
                AddendaRecord(
                    self.entry_detail.std_ent_cls_code,
                    pmt_rel_info=addenda.get("payment_related_info").upper(),
                    add_seq_num=index + 1,
                    ent_det_seq_num=entry_detail.trace_num[-7:],
                )
            )

        if self.addenda_record:
            self.entry_detail.add_rec_ind = 1

    def render_to_string(self, force_crlf=False):
        """
        Renders a nacha batch entry and addenda to string
        """
        line_ending = "\n"
        if force_crlf:
            line_ending = "\r\n"

        ret_string = self.entry_detail.get_row() + line_ending

        for addenda in self.addenda_record:
            ret_string += addenda.get_row() + line_ending

        return ret_string